ImplicitReplyActivity subclass: #EnabledSelectorsActivity
	instanceVariableNames: 'enabledSelectors '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-SelectActivity'!
EnabledSelectorsActivity comment:
'Abstract class EnabledSelectorsActivity provides control over acceptance of messages through a set of enabled methods.
Example of use if for implementing abstract states (category Actalk-Synchro-State).'!


!EnabledSelectorsActivity methodsFor: 'message handling'!

nextMessage
	"Look into the mailbox for the first message matching one of the enabled selectors."

	"Note that this is an abstract class.
	The subclass needs to define the actual assignment of the set of enabled selectors."

	^self mailBox firstMessageWithCondition: [:message |
		enabledSelectors includes: message selector]! !

ImplicitReplyActivity subclass: #SelectiveAcceptActivity
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-SelectActivity'!
SelectiveAcceptActivity comment:
'Abstract class SelectiveAcceptActivity provides control over acceptance of messages by explicit check (method isAcceptableMessage:). Rejected messages are resent.
Example of use if for naive implementation of guards (category Actalk-Synchro-Guards).'!


!SelectiveAcceptActivity methodsFor: 'message handling'!

isAcceptableMessage: aMessage
	"Checking if a message may be accepted."
	"Default check is always true."

	^true!

nextMessage
	"Next message is first acceptable message.
	Resend other ones."

	| message |
	message := super nextMessage.
	[self isAcceptableMessage: message] whileFalse:
		[address resend: message.
		message := super nextMessage].
	^message! !
